import sys
import time

from twisted.internet import reactor
from twisted.internet.defer import (
    returnValue,
    inlineCallbacks,
    maybeDeferred as _maybeDeferred,
    DeferredList,
    Deferred)
from twisted.python.failure import Failure


def maybeDeferred(f, *args, **kwargs):
    """Wrapper to Twisted's maybeDeferred.

    Adds a reference to the wrapped deferred instance inside the original
    function dict so we can track if something is a deferred.
    """
    deferred = _maybeDeferred(f, *args, **kwargs)
    f.func_dict['deferred'] = deferred
    return deferred


def maybeDeferToThread(f, *args, **kwargs):
    """Call the function C{f} using a thread from the given threadpool

    Return sthe result as a Deferred.

    @param f: The function to call. May return a deferred.
    @param *args: positional arguments to pass to f.
    @param **kwargs: keyword arguments to pass to f.

    @return: A Deferred which fires a callback with the result of f, or an
        errback with a L{twisted.python.failure.Failure} if f throws an
        exception.
    """
    threadpool = reactor.getThreadPool()

    d = Deferred()

    def realOnResult(result):
        if not isinstance(result, Failure):
            reactor.callFromThread(d.callback, result)
        else:
            reactor.callFromThread(d.errback, result)

    def onResult(success, result):
        assert success
        assert isinstance(result, Deferred)
        result.addBoth(realOnResult)

    threadpool.callInThreadWithCallback(onResult, maybeDeferred,
                                        f, *args, **kwargs)

    return d


class Check(object):
    """Abstract base class for objects embodying connectivity checks."""

    def check(self, pattern, results):
        """Run this check, if it matches the pattern.

        If the pattern matches, and this is a leaf node in the check tree,
        implementations of Check.check should call
        results.notify_start, then either results.notify_success or
        results.notify_failure.
        """
        raise NotImplementedError("{}.check not implemented".format(
                                  type(self)))

    def skip(self, pattern, results):
        """Indicate that this check has been skipped.

        If the pattern matches and this is a leaf node in the check tree,
        implementations of Check.skip should call results.notify_skip.
        """
        raise NotImplementedError("{}.skip not implemented".format(
                                  type(self)))


class ConditionalCheck(Check):
    """A Check that skips unless the given predicate is true at check time."""

    def __init__(self, wrapped, predicate):
        """Initialize an instance."""
        super(ConditionalCheck, self).__init__()
        self.wrapped = wrapped
        self.predicate = predicate

    def check(self, pattern, result):
        """Skip the check."""
        if self.predicate():
            return self.wrapped.check(pattern, result)
        else:
            self.skip(pattern, result)

    def skip(self, pattern, result):
        """Skip the check."""
        self.wrapped.skip(pattern, result)


class ResultTracker(object):
    """Base class for objects which report or record check results."""

    def notify_start(self, name, info):
        """Register the start of a check."""

    def notify_skip(self, name):
        """Register a check being skipped."""

    def notify_success(self, name, duration):
        """Register a successful check."""

    def notify_failure(self, name, info, exc_info, duration):
        """Register the failure of a check."""


class PrefixResultWrapper(ResultTracker):
    """ResultWrapper wrapper which adds a prefix to recorded results."""

    def __init__(self, wrapped, prefix):
        """Initialize an instance."""
        super(PrefixResultWrapper, self).__init__()
        self.wrapped = wrapped
        self.prefix = prefix

    def make_name(self, name):
        """Make a name by prepending the prefix."""
        return "{}{}".format(self.prefix, name)

    def notify_skip(self, name):
        """Register a check being skipped."""
        self.wrapped.notify_skip(self.make_name(name))

    def notify_start(self, name, info):
        """Register the start of a check."""
        self.wrapped.notify_start(self.make_name(name), info)

    def notify_success(self, name, duration):
        """Register success."""
        self.wrapped.notify_success(self.make_name(name), duration)

    def notify_failure(self, name, info, exc_info, duration):
        """Register failure."""
        self.wrapped.notify_failure(self.make_name(name),
                                    info, exc_info, duration)


class FailureCountingResultWrapper(ResultTracker):
    """ResultWrapper wrapper which counts failures."""

    def __init__(self, wrapped):
        """Initialize an instance."""
        super(FailureCountingResultWrapper, self).__init__()
        self.wrapped = wrapped
        self.failure_count = 0

    def notify_skip(self, name):
        """Register a check being skipped."""
        self.wrapped.notify_skip(name)

    def notify_start(self, name, info):
        """Register the start of a check."""
        self.failure_count += 1
        self.wrapped.notify_start(name, info)

    def notify_success(self, name, duration):
        """Register success."""
        self.failure_count -= 1
        self.wrapped.notify_success(name, duration)

    def notify_failure(self, name, info, exc_info, duration):
        """Register failure."""
        self.wrapped.notify_failure(name, info, exc_info, duration)

    def any_failed(self):
        """Return True if any checks using this wrapper failed so far."""
        return self.failure_count > 0


class FunctionCheck(Check):
    """A Check which takes a check function."""

    def __init__(self, name, check, info=None, blocking=False):
        """Initialize an instance."""
        super(FunctionCheck, self).__init__()
        self.name = name
        self.info = info
        self.check_fn = check
        self.blocking = blocking

    @inlineCallbacks
    def check(self, pattern, results):
        """Call the check function."""
        if not pattern.matches(self.name):
            returnValue(None)
        results.notify_start(self.name, self.info)
        start = time.time()
        try:
            if self.blocking:
                result = yield maybeDeferToThread(self.check_fn)
            else:
                result = yield maybeDeferred(self.check_fn)
            results.notify_success(self.name, time.time() - start)
            returnValue(result)
        except Exception:
            results.notify_failure(self.name, self.info,
                                   sys.exc_info(), time.time() - start)

    def skip(self, pattern, results):
        """Record the skip."""
        if not pattern.matches(self.name):
            return
        results.notify_skip(self.name)


class MultiCheck(Check):
    """A composite check comprised of multiple subchecks."""

    def __init__(self, subchecks, strategy):
        """Initialize an instance."""
        super(MultiCheck, self).__init__()
        self.subchecks = list(subchecks)
        self.strategy = strategy

    def check(self, pattern, results):
        """Run subchecks using the strategy supplied at creation time."""
        return self.strategy(self.subchecks, pattern, results)

    def skip(self, pattern, results):
        """Skip subchecks."""
        for subcheck in self.subchecks:
            subcheck.skip(pattern, results)


class PrefixCheckWrapper(Check):
    """Runs a given check, adding a prefix to its name.

    This works by wrapping the pattern and result tracker objects
    passed to .check and .skip.
    """

    def __init__(self, wrapped, prefix):
        """Initialize an instance."""
        super(PrefixCheckWrapper, self).__init__()
        self.wrapped = wrapped
        self.prefix = prefix

    def do_subcheck(self, subcheck, pattern, results):
        """Do a subcheck if the pattern could still match."""
        pattern = pattern.assume_prefix(self.prefix)
        if not pattern.failed():
            results = PrefixResultWrapper(wrapped=results,
                                          prefix=self.prefix)
            return subcheck(pattern, results)

    def check(self, pattern, results):
        """Run the check, prefixing results."""
        return self.do_subcheck(self.wrapped.check, pattern, results)

    def skip(self, pattern, results):
        """Skip checks, prefixing results."""
        self.do_subcheck(self.wrapped.skip, pattern, results)


@inlineCallbacks
def skipping_strategy(subchecks, pattern, results):
    """Strategy used to print checks out by just skipping them all."""
    # We need at least one yield to make this a generator
    yield
    for subcheck in subchecks:
        subcheck.skip(pattern, results)


@inlineCallbacks
def sequential_strategy(subchecks, pattern, results):
    """Run subchecks sequentially, skipping checks after the first failure.

    This is most useful when the failure of one check in the sequence
    would imply the failure of later checks -- for example, it probably
    doesn't make sense to run an SSL check if the basic TCP check failed.

    Use sequential_check to create a meta-check using this strategy.
    """
    local_results = FailureCountingResultWrapper(wrapped=results)
    failed = False
    for subcheck in subchecks:
        if failed:
            subcheck.skip(pattern, local_results)
        else:
            yield maybeDeferred(subcheck.check, pattern, local_results)
            if local_results.any_failed():
                failed = True


def parallel_strategy(subchecks, pattern, results):
    """A strategy which runs the given subchecks in parallel.

    Most checks can potentially block for long periods, and shouldn't have
    interdependencies, so it makes sense to run them in parallel to
    shorten the overall run time.

    Use parallel_check to create a meta-check using this strategy.
    """
    deferreds = [maybeDeferred(subcheck.check, pattern, results)
                 for subcheck in subchecks]
    return DeferredList(deferreds)


def skipping_check(subchecks):
    """Return a check that skips everything, used for printing checks."""
    return MultiCheck(subchecks=subchecks, strategy=skipping_strategy)


def parallel_check(subchecks):
    """Return a check that runs the given subchecks in parallel."""
    return MultiCheck(subchecks=subchecks, strategy=parallel_strategy)


def sequential_check(subchecks):
    """Return a check that runs the given subchecks in sequence."""
    return MultiCheck(subchecks=subchecks, strategy=sequential_strategy)


def add_check_prefix(*args):
    """Return an equivalent check with the given prefix prepended to its name.

    The final argument should be a check; the remaining arguments are treated
    as name components and joined with the check name using periods as
    separators.  For example, if the name of a check is "baz", then:

        add_check_prefix("foo", "bar", check)

    ...will return a check with the effective name "foo.bar.baz".
    """
    args = list(args)
    check = args.pop(-1)
    path = ".".join(args)
    return PrefixCheckWrapper(wrapped=check, prefix="{}:".format(path))


def make_check(name, check, info=None, blocking=False):
    """Make a check object from a function."""
    return FunctionCheck(name=name, check=check, info=info, blocking=blocking)


def guard_check(check, predicate):
    """Wrap a check so that it is skipped unless the predicate is true."""
    return ConditionalCheck(wrapped=check, predicate=predicate)



import operator
import random
import testtools
from StringIO import StringIO

from testtools import matchers

from conn_check.check_impl import (
    FunctionCheck,
    MultiCheck,
    parallel_strategy,
    PrefixCheckWrapper,
    sequential_strategy,
    )
from conn_check.checks import (
    CHECKS,
    extract_host_port,
    make_amqp_check,
    make_http_check,
    make_memcache_check,
    make_mongodb_check,
    make_postgres_check,
    make_redis_check,
    make_smtp_check,
    make_tls_check,
    make_tcp_check,
    make_udp_check,
    )
from conn_check.main import (
    build_checks,
    check_from_description,
    OrderedOutput,
    )


class FunctionCheckMatcher(testtools.Matcher):

    def __init__(self, name, info, blocking=False):
        self.name = name
        self.info = info
        self.blocking = blocking

    def match(self, matchee):
        checks = []
        checks.append(matchers.IsInstance(FunctionCheck))
        checks.append(matchers.Annotate(
            "name doesn't match",
            matchers.AfterPreprocessing(operator.attrgetter('name'),
                matchers.Equals(self.name))))
        checks.append(matchers.Annotate(
            "info doesn't match",
            matchers.AfterPreprocessing(operator.attrgetter('info'),
                matchers.Equals(self.info))))
        checks.append(matchers.Annotate(
            "blocking doesn't match",
            matchers.AfterPreprocessing(operator.attrgetter('blocking'),
                matchers.Equals(self.blocking))))
        return matchers.MatchesAll(*checks).match(matchee)

    def __str__(self):
        return ("Is a FunctionCheck with <name={} info={} "
                "blocking={}>".format(self.name, self.info, self.blocking))


class MultiCheckMatcher(testtools.Matcher):

    def __init__(self, strategy, subchecks):
        self.strategy = strategy
        self.subchecks = subchecks

    def match(self, matchee):
        checks = []
        checks.append(matchers.IsInstance(MultiCheck))
        checks.append(matchers.AfterPreprocessing(operator.attrgetter('strategy'),
                        matchers.Is(self.strategy)))
        checks.append(matchers.AfterPreprocessing(operator.attrgetter('subchecks'),
                        matchers.MatchesListwise(self.subchecks)))
        return matchers.MatchesAll(*checks).match(matchee)

    def __str__(self):
        return ("Is a MultiCheck with <strategy={} subchecks={}>"
                "".format(self.strategy, self.subchecks))


class ExtractHostPortTests(testtools.TestCase):

    def test_basic(self):
        self.assertEqual(extract_host_port('http://localhost:80/'),
            ('localhost', 80, 'http'))

    def test_no_scheme(self):
        self.assertEqual(extract_host_port('//localhost/'),
            ('localhost', 80, 'http'))

    def test_no_port_http(self):
        self.assertEqual(extract_host_port('http://localhost/'),
            ('localhost', 80, 'http'))

    def test_no_port_https(self):
        self.assertEqual(extract_host_port('https://localhost/'),
            ('localhost', 443, 'https'))


class ConnCheckTest(testtools.TestCase):

    def test_make_tcp_check(self):
        result = make_tcp_check('localhost', 8080)
        self.assertThat(result, FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))

    def test_make_tls_check(self):
        result = make_tls_check('localhost', 8080, verify=True)
        self.assertThat(result, FunctionCheckMatcher('tls:localhost:8080', 'localhost:8080'))

    def test_make_udp_check(self):
        result = make_udp_check('localhost', 8080, 'foo', 'bar')
        self.assertThat(result, FunctionCheckMatcher('udp:localhost:8080', 'localhost:8080'))

    def test_make_http_check(self):
        result = make_http_check('http://localhost/')
        self.assertIsInstance(result, PrefixCheckWrapper)
        self.assertEqual(result.prefix, 'http:http://localhost/:')
        wrapped = result.wrapped
        self.assertIsInstance(wrapped, MultiCheck)
        self.assertIs(wrapped.strategy, sequential_strategy)
        self.assertEqual(len(wrapped.subchecks), 2)
        self.assertThat(wrapped.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:80', 'localhost:80'))
        self.assertThat(wrapped.subchecks[1],
                FunctionCheckMatcher('', 'GET http://localhost/'))

    def test_make_http_check_https(self):
        result = make_http_check('https://localhost/')
        self.assertIsInstance(result, PrefixCheckWrapper)
        self.assertEqual(result.prefix, 'http:https://localhost/:')
        wrapped = result.wrapped
        self.assertIsInstance(wrapped, MultiCheck)
        self.assertIs(wrapped.strategy, sequential_strategy)
        self.assertEqual(len(wrapped.subchecks), 2)
        self.assertThat(wrapped.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:443', 'localhost:443'))
        self.assertThat(wrapped.subchecks[1],
                FunctionCheckMatcher('', 'GET https://localhost/'))

    def test_make_amqp_check(self):
        result = make_amqp_check('localhost', 8080, 'foo',
                                 'bar', use_tls=True, vhost='/')
        self.assertIsInstance(result, MultiCheck)
        self.assertIs(result.strategy, sequential_strategy)
        self.assertEqual(len(result.subchecks), 3)
        self.assertThat(result.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(result.subchecks[1],
                FunctionCheckMatcher('tls:localhost:8080', 'localhost:8080'))
        self.assertThat(result.subchecks[2],
                FunctionCheckMatcher('amqp:localhost:8080', 'user foo'))

    def test_make_amqp_check_no_tls(self):
        result = make_amqp_check('localhost', 8080, 'foo',
                                 'bar', use_tls=False, vhost='/')
        self.assertIsInstance(result, MultiCheck)
        self.assertIs(result.strategy, sequential_strategy)
        self.assertEqual(len(result.subchecks), 2)
        self.assertThat(result.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(result.subchecks[1],
                FunctionCheckMatcher('amqp:localhost:8080', 'user foo'))

    def test_make_postgres_check(self):
        result = make_postgres_check('localhost', 8080,'foo',
                                     'bar', 'test')
        self.assertIsInstance(result, MultiCheck)
        self.assertIs(result.strategy, sequential_strategy)
        self.assertEqual(len(result.subchecks), 2)
        self.assertThat(result.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(result.subchecks[1],
                FunctionCheckMatcher('postgres:localhost:8080', 'user foo', blocking=True))

    def test_make_postgres_check_local_socket(self):
        result = make_postgres_check('/local.sock', 8080,'foo',
                                     'bar', 'test')
        self.assertIsInstance(result, MultiCheck)
        self.assertIs(result.strategy, sequential_strategy)
        self.assertEqual(len(result.subchecks), 1)
        self.assertThat(result.subchecks[0],
                FunctionCheckMatcher('postgres:/local.sock:8080', 'user foo', blocking=True))

    def test_make_redis_check(self):
        result = make_redis_check('localhost', 8080)
        self.assertIsInstance(result, PrefixCheckWrapper)
        self.assertEqual(result.prefix, 'redis:localhost:8080:')
        wrapped = result.wrapped
        self.assertIsInstance(wrapped, MultiCheck)
        self.assertIs(wrapped.strategy, sequential_strategy)
        self.assertEqual(len(wrapped.subchecks), 2)
        self.assertThat(wrapped.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(wrapped.subchecks[1], FunctionCheckMatcher('connect', None))

    def test_make_redis_check_with_password(self):
        result = make_redis_check('localhost', 8080, 'foobar')
        self.assertIsInstance(result, PrefixCheckWrapper)
        self.assertEqual(result.prefix, 'redis:localhost:8080:')
        wrapped = result.wrapped
        self.assertIsInstance(wrapped, MultiCheck)
        self.assertIs(wrapped.strategy, sequential_strategy)
        self.assertEqual(len(wrapped.subchecks), 2)
        self.assertThat(wrapped.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(wrapped.subchecks[1],
                        FunctionCheckMatcher('connect with auth', None))

    def test_make_memcache_check(self):
        result = make_memcache_check('localhost', 8080)
        self.assertIsInstance(result, PrefixCheckWrapper)
        self.assertEqual(result.prefix, 'memcache:localhost:8080:')
        wrapped = result.wrapped
        self.assertIsInstance(wrapped, MultiCheck)
        self.assertIs(wrapped.strategy, sequential_strategy)
        self.assertEqual(len(wrapped.subchecks), 2)
        self.assertThat(wrapped.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(wrapped.subchecks[1], FunctionCheckMatcher('connect', None))

    def test_make_mongodb_check(self):
        result = make_mongodb_check('localhost', 8080)
        self.assertIsInstance(result, PrefixCheckWrapper)
        self.assertEqual(result.prefix, 'mongodb:localhost:8080:')
        wrapped = result.wrapped
        self.assertIsInstance(wrapped, MultiCheck)
        self.assertIs(wrapped.strategy, sequential_strategy)
        self.assertEqual(len(wrapped.subchecks), 2)
        self.assertThat(wrapped.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(wrapped.subchecks[1], FunctionCheckMatcher('connect', None))

    def test_make_mongodb_check_with_username(self):
        result = make_mongodb_check('localhost', 8080, 'foo')
        self.assertIsInstance(result, PrefixCheckWrapper)
        self.assertEqual(result.prefix, 'mongodb:localhost:8080:')
        wrapped = result.wrapped
        self.assertIsInstance(wrapped, MultiCheck)
        self.assertIs(wrapped.strategy, sequential_strategy)
        self.assertEqual(len(wrapped.subchecks), 2)
        self.assertThat(wrapped.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(wrapped.subchecks[1],
                FunctionCheckMatcher('connect with auth', None))

    def test_make_smtp_check(self):
        result = make_smtp_check('localhost', 8080, 'foo', 'bar',
                                 'foo@example.com', 'bax@example.com',
                                 use_tls=True)
        self.assertIsInstance(result, MultiCheck)
        self.assertIs(result.strategy, sequential_strategy)
        self.assertEqual(len(result.subchecks), 3)
        self.assertThat(result.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(result.subchecks[1],
                FunctionCheckMatcher('tls:localhost:8080', 'localhost:8080'))
        self.assertThat(result.subchecks[2],
                FunctionCheckMatcher('smtp:localhost:8080', 'user foo'))

    def test_make_smtp_check_no_tls(self):
        result = make_smtp_check('localhost', 8080, 'foo', 'bar',
                                 'foo@example.com', 'bax@example.com',
                                 use_tls=False)
        self.assertIsInstance(result, MultiCheck)
        self.assertIs(result.strategy, sequential_strategy)
        self.assertEqual(len(result.subchecks), 2)
        self.assertThat(result.subchecks[0],
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))
        self.assertThat(result.subchecks[1],
                FunctionCheckMatcher('smtp:localhost:8080', 'user foo'))

    def test_check_from_description_unknown_type(self):
        e = self.assertRaises(AssertionError,
                              check_from_description, {'type': 'foo'})
        self.assertEqual(
            str(e),
            "Unknown check type: foo, available checks: {}".format(CHECKS.keys()))

    def test_check_from_description_missing_arg(self):
        description = {'type': 'tcp'}
        e = self.assertRaises(AssertionError,
                check_from_description, description)
        self.assertEqual(
            str(e),
            "host missing from check: {}".format(description))

    def test_check_from_description_makes_check(self):
        description = {'type': 'tcp', 'host': 'localhost', 'port': '8080'}
        result = check_from_description(description)
        self.assertThat(result,
                FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'))

    def test_build_checks(self):
        description = [{'type': 'tcp', 'host': 'localhost', 'port': '8080'}]
        result = build_checks(description, 10, [], [])
        self.assertThat(result,
                MultiCheckMatcher(strategy=parallel_strategy,
                    subchecks=[FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080')]))

    def test_build_checks_with_tags(self):
        descriptions = [
            {'type': 'tcp', 'host': 'localhost', 'port': '8080'},
            {'type': 'tcp', 'host': 'localhost2', 'port': '8080',
             'tags': ['foo']},
            {'type': 'tcp', 'host': 'localhost3', 'port': '8080',
             'tags': ['foo', 'bar']},
            {'type': 'tcp', 'host': 'localhost4', 'port': '8080',
             'tags': ['baz']},
            {'type': 'tcp', 'host': 'localhost5', 'port': '8080',
             'tags': ['bar']},
        ]
        result = build_checks(descriptions, 10, ['foo', 'bar'], [])
        expected_subchecks = [
            FunctionCheckMatcher('tcp:localhost2:8080', 'localhost2:8080'),
            FunctionCheckMatcher('tcp:localhost3:8080', 'localhost3:8080'),
            FunctionCheckMatcher('tcp:localhost5:8080', 'localhost5:8080'),
        ]
        self.assertThat(result,
                MultiCheckMatcher(strategy=parallel_strategy,
                    subchecks=expected_subchecks))

    def test_build_checks_with_excluded_tags(self):
        descriptions = [
            {'type': 'tcp', 'host': 'localhost', 'port': '8080'},
            {'type': 'tcp', 'host': 'localhost2', 'port': '8080',
             'tags': ['foo']},
            {'type': 'tcp', 'host': 'localhost3', 'port': '8080',
             'tags': ['foo', 'bar']},
            {'type': 'tcp', 'host': 'localhost4', 'port': '8080',
             'tags': ['baz']},
            {'type': 'tcp', 'host': 'localhost5', 'port': '8080',
             'tags': ['bar']},
        ]
        result = build_checks(descriptions, 10, [], ['bar', 'baz'])
        expected_subchecks = [
            FunctionCheckMatcher('tcp:localhost:8080', 'localhost:8080'),
            FunctionCheckMatcher('tcp:localhost2:8080', 'localhost2:8080'),
        ]
        self.assertThat(result,
                MultiCheckMatcher(strategy=parallel_strategy,
                    subchecks=expected_subchecks))

    def test_ordered_output(self):
        lines = [
            'SKIPPED: xyz3:localhost:666\n',
            'bar2:localhost:8080 FAILED: error\n',
            'SKIPPED: foo2:localhost:8080\n',
            'baz2:localhost:42 OK\n',
            'SKIPPED: bar2:localhost:8080\n',
            'xyz2:localhost:666 FAILED: error\n',
            'xyz1:localhost:666 OK\n',
            'foo1:localhost:8080 FAILED: error\n',
            'baz1:localhost:42 OK\n',
        ]
        expected = (
            'bar2:localhost:8080 FAILED: error\n'
            'foo1:localhost:8080 FAILED: error\n'
            'xyz2:localhost:666 FAILED: error\n'
            'baz1:localhost:42 OK\n'
            'baz2:localhost:42 OK\n'
            'xyz1:localhost:666 OK\n'
            'SKIPPED: bar2:localhost:8080\n'
            'SKIPPED: foo2:localhost:8080\n'
            'SKIPPED: xyz3:localhost:666\n'
        )

        output = OrderedOutput(StringIO())
        map(output.write, lines)
        output.flush()
        self.assertEqual(expected, output.output.getvalue())

        output = OrderedOutput(StringIO())
        random.shuffle(lines)
        map(output.write, lines)
        output.flush()
        self.assertEqual(expected, output.output.getvalue())



from argparse import ArgumentParser
from collections import defaultdict
import sys
from threading import Thread
import time
import traceback
import yaml

from twisted.internet import reactor
from twisted.internet.defer import (
    inlineCallbacks,
    )
from twisted.python.threadpool import ThreadPool

from . import get_version_string
from .check_impl import (
    FailureCountingResultWrapper,
    parallel_check,
    skipping_check,
    ResultTracker,
    )
from .checks import CHECK_ALIASES, CHECKS, load_tls_certs
from .patterns import (
    SimplePattern,
    SumPattern,
    )


def check_from_description(check_description):
    _type = check_description['type']

    if _type in CHECK_ALIASES:
        _type = CHECK_ALIASES[_type]

    check = CHECKS.get(_type, None)
    if check is None:
        raise AssertionError("Unknown check type: {}, available checks: {}"
                             .format(_type, CHECKS.keys()))
    for arg in check['args']:
        if arg not in check_description:
            raise AssertionError('{} missing from check: {}'.format(arg,
                                 check_description))

    if 'port' in check_description:
        check_description['port'] = int(check_description['port'])

    res = check['fn'](**check_description)
    return res


def filter_tags(check, include, exclude):
    if not include and not exclude:
        return True

    check_tags = set(check.get('tags', []))

    if include:
        result = bool(check_tags.intersection(include))
    else:
        result = not bool(check_tags.intersection(exclude))

    return result


def build_checks(check_descriptions, connect_timeout, include_tags,
                 exclude_tags, skip_checks=False):
    def set_timeout(desc):
        new_desc = dict(timeout=connect_timeout)
        new_desc.update(desc)
        return new_desc

    check_descriptions = filter(
        lambda c: filter_tags(c, include_tags, exclude_tags),
        check_descriptions)

    subchecks = map(
        lambda c: check_from_description(c),
        map(set_timeout, check_descriptions))

    if skip_checks:
        strategy_wrapper = skipping_check
    else:
        strategy_wrapper = parallel_check
    return strategy_wrapper(subchecks)


@inlineCallbacks
def run_checks(checks, pattern, results):
    """Make and run all the pertinent checks."""
    try:
        yield checks.check(pattern, results)
    finally:
        reactor.stop()


class NagiosCompatibleArgsParser(ArgumentParser):

    def error(self, message):
        """A patched version of ArgumentParser.error.

        Does the same thing as ArgumentParser.error, e.g. prints an error
        message and exits, but does so with an exit code of 3 rather than 2,
        to maintain compatibility with Nagios checks.
        """
        self.print_usage(sys.stderr)
        self.exit(3, '{}: error: {}\n'.format(self.prog, message))


class TimestampOutput(object):

    def __init__(self, output):
        self.start = time.time()
        self.output = output

    def write(self, data):
        self.output.write("{:.3f}: {}".format(time.time() - self.start, data))


class OrderedOutput(object):
    """Outputs check results ordered by FAILED, SUCCESSFUL, SKIPPED checks."""

    def __init__(self, output):
        self.output = output

        self.failed = defaultdict(list)
        self.messages = defaultdict(list)
        self.skipped = []

    def write(self, data):
        if data[:7] == 'SKIPPED':
            self.skipped.append(data)
            return

        name, message = data.split(' ', 1)

        # Standard check name format is {type}:{host}:{port}
        name_parts = name.split(':', 2)
        try:
            name_parts[2] = ''
        except IndexError:
            pass
        name = ':'.join(name_parts)

        if message[0:6] == 'FAILED':
            self.failed[name].append(data)
        else:
            self.messages[name].append(data)

    def flush(self):
        for _type in ('failed', 'messages'):
            for name, messages in sorted(getattr(self, _type).items()):
                messages.sort()
                map(self.output.write, messages)

        self.skipped.sort()
        map(self.output.write, self.skipped)


class ConsoleOutput(ResultTracker):
    """Outputs check results to STDOUT."""

    def __init__(self, output, verbose, show_tracebacks, show_duration):
        """Initialize an instance."""
        super(ConsoleOutput, self).__init__()
        self.output = output
        self.verbose = verbose
        self.show_tracebacks = show_tracebacks
        self.show_duration = show_duration

    def format_duration(self, duration):
        if not self.show_duration:
            return ""
        return ": ({:.3f} ms)".format(duration)

    def notify_start(self, name, info):
        """Register the start of a check."""
        if self.verbose:
            if info:
                info = " ({})".format(info)
            else:
                info = ''
            self.output.write("Starting {}{}...\n".format(name, info))

    def notify_skip(self, name):
        """Register a check being skipped."""
        self.output.write("SKIPPED: {}\n".format(name))

    def notify_success(self, name, duration):
        """Register a success."""
        self.output.write("{} OK{}\n".format(
            name, self.format_duration(duration)))

    def notify_failure(self, name, info, exc_info, duration):
        """Register a failure."""
        message = str(exc_info[1]).split("\n")[0]
        if info:
            message = "({}) {}".format(info, message)
        self.output.write("{} FAILED{} - {}\n".format(
            name, self.format_duration(duration), message))

        if self.show_tracebacks:
            formatted = traceback.format_exception(exc_info[0],
                                                   exc_info[1],
                                                   exc_info[2],
                                                   None)
            lines = "".join(formatted).split("\n")
            if len(lines) > 0 and len(lines[-1]) == 0:
                lines.pop()
            indented = "\n".join(["  {}".format(line) for line in lines])
            self.output.write("{}\n".format(indented))


class Command(object):
    """CLI command runner for the main conn-check endpoint."""

    def __init__(self, args):
        self.make_arg_parser()
        self.parse_options(args)
        self.wrap_output(sys.stdout)
        self.load_descriptions()

    def make_arg_parser(self):
        """Set up an arg parser with our options."""

        parser = NagiosCompatibleArgsParser()
        parser.add_argument("config_file",
                            help="Config file specifying the checks to run.")
        parser.add_argument("patterns", nargs='*',
                            help="Patterns to filter the checks.")
        parser.add_argument("-v", "--verbose", dest="verbose",
                            action="store_true", default=False,
                            help="Show additional status")
        parser.add_argument("-d", "--duration", dest="show_duration",
                            action="store_true", default=False,
                            help="Show duration")
        parser.add_argument("-t", "--tracebacks", dest="show_tracebacks",
                            action="store_true", default=False,
                            help="Show tracebacks on failure")
        parser.add_argument("--validate", dest="validate",
                            action="store_true", default=False,
                            help="Only validate the config file,"
                            " don't run checks.")
        parser.add_argument("--version", dest="print_version",
                            action="store_true", default=False,
                            help="Print the currently installed version.")
        parser.add_argument("--tls-certs-path", dest="cacerts_path",
                            action="store", default="/etc/ssl/certs/",
                            help="Path to TLS CA certificates.")
        parser.add_argument("--max-timeout", dest="max_timeout", type=float,
                            action="store", help="Maximum execution time.")
        parser.add_argument("--connect-timeout", dest="connect_timeout",
                            action="store", default=10, type=float,
                            help="Network connection timeout.")
        parser.add_argument("-U", "--unbuffered-output", dest="buffer_output",
                            action="store_false", default=True,
                            help="Don't buffer output, write to STDOUT right "
                            "away.")
        parser.add_argument("--dry-run",
                            dest="dry_run", action="store_true",
                            default=False,
                            help="Skip all checks, just print out"
                            " what would be run.")
        group = parser.add_mutually_exclusive_group()
        group.add_argument("--include-tags", dest="include_tags",
                           action="store", default="",
                           help="Comma separated list of tags to include.")
        group.add_argument("--exclude-tags", dest="exclude_tags",
                           action="store", default="",
                           help="Comma separated list of tags to exclude.")
        self.parser = parser

    def setup_reactor(self):
        """Setup the Twisted reactor with required customisations."""

        def make_daemon_thread(*args, **kw):
            """Create a daemon thread."""
            thread = Thread(*args, **kw)
            thread.daemon = True
            return thread

        threadpool = ThreadPool(minthreads=1)
        threadpool.threadFactory = make_daemon_thread
        reactor.threadpool = threadpool
        reactor.callWhenRunning(threadpool.start)

        if self.options.max_timeout is not None:
            def terminator():
                # Hasta la vista, twisted
                reactor.stop()
                print('Maximum timeout reached: {}s'.format(
                      self.options.max_timeout))

            reactor.callLater(self.options.max_timeout, terminator)

    def parse_options(self, args):
        """Parse args (e.g. sys.argv) into options and set some config."""

        options = self.parser.parse_args(list(args))

        include_tags = []
        if options.include_tags:
            include_tags = options.include_tags.split(',')
            include_tags = [tag.strip() for tag in include_tags]
        options.include_tags = include_tags

        exclude_tags = []
        if options.exclude_tags:
            exclude_tags = options.exclude_tags.split(',')
            exclude_tags = [tag.strip() for tag in exclude_tags]
        options.exclude_tags = exclude_tags

        if options.patterns:
            self.patterns = SumPattern(map(SimplePattern, options.patterns))
        else:
            self.patterns = SimplePattern("*")
        self.options = options

    def wrap_output(self, output):
        """Wraps an output stream (e.g. sys.stdout) from options."""

        if self.options.show_duration:
            output = TimestampOutput(output)
        if self.options.buffer_output:
            # We buffer output so we can order it for human readable output
            output = OrderedOutput(output)

        results = ConsoleOutput(output=output,
                                show_tracebacks=self.options.show_tracebacks,
                                show_duration=self.options.show_duration,
                                verbose=self.options.verbose)
        if not self.options.dry_run:
            results = FailureCountingResultWrapper(results)

        self.output = output
        self.results = results

    def load_descriptions(self):
        """Pre-load YAML checks file into a descriptions property."""

        with open(self.options.config_file) as f:
            self.descriptions = yaml.load(f)

    def run(self):
        """Run/validate/dry-run the given command with options."""

        checks = build_checks(self.descriptions,
                              self.options.connect_timeout,
                              self.options.include_tags,
                              self.options.exclude_tags,
                              self.options.dry_run)

        if not self.options.validate:
            if not self.options.dry_run:
                load_tls_certs(self.options.cacerts_path)

            self.setup_reactor()
            reactor.callWhenRunning(run_checks, checks, self.patterns,
                                    self.results)
            reactor.run()

            # Flush output, this really only has an effect when running
            # buffered output
            self.output.flush()

            if not self.options.dry_run and self.results.any_failed():
                return 2

        return 0


def parse_version_arg():
    """Manually check for --version in args and output version info.

    We need to do this early because ArgumentParser won't let us mix
    and match non-default positional argument with a flag argument.
    """
    if '--version' in sys.argv:
        sys.stdout.write('conn-check {}\n'.format(get_version_string()))
        return True


def run(*args):
    if parse_version_arg():
        return 0

    cmd = Command(args)
    return cmd.run()


def main():
    sys.exit(run(*sys.argv[1:]))


if __name__ == '__main__':
    main()



import re


class Pattern(object):
    """Abstract base class for patterns used to select subsets of checks."""

    def assume_prefix(self, prefix):
        """Return an equivalent pattern with the given prefix baked in.

        For example, if self.matches("bar") is True, then
        self.assume_prefix("foo").matches("foobar") will be True.
        """
        return PrefixPattern(prefix, self)

    def failed(self):
        """Return True if the pattern cannot match any string.

        This is mainly used so we can bail out early when recursing into
        check trees.
        """
        return not self.prefix_matches("")

    def prefix_matches(self, partial_name):
        """Return True if the partial name (a prefix) is a potential match."""
        raise NotImplementedError("{}.prefix_matches not "
                                  "implemented".format(type(self)))

    def matches(self, name):
        """Return True if the given name matches."""
        raise NotImplementedError("{}.match not "
                                  "implemented".format(type(self)))


class FailedPattern(Pattern):
    """Patterns that always fail to match."""

    def assume_prefix(self, prefix):
        """Return an equivalent pattern with the given prefix baked in."""
        return FAILED_PATTERN

    def prefix_matches(self, partial_name):
        """Return True if the partial name matches."""
        return False

    def matches(self, name):
        """Return True if the complete name matches."""
        return False


FAILED_PATTERN = FailedPattern()


PATTERN_TOKEN_RE = re.compile(r'\*|[^*]+')


def tokens_to_partial_re(tokens):
    """Convert tokens to a regular expression for matching prefixes."""

    def token_to_re(token):
        """Convert tokens to (begin, end, alt_end) triples."""
        if token == '*':
            return (r'(?:.*', ')?', ')')
        else:
            chars = list(token)
            begin = "".join(["(?:" + re.escape(c) for c in chars])
            end = "".join([")?" for c in chars])
            return (begin, end, end)

    subexprs = map(token_to_re, tokens)
    if len(subexprs) > 0:
        # subexpressions like (.*)? aren't accepted, so we may have to use
        # an alternate closing form for the last (innermost) subexpression
        (begin, _, alt_end) = subexprs[-1]
        subexprs[-1] = (begin, alt_end, alt_end)
    return re.compile("".join([se[0] for se in subexprs] +
                              [se[1] for se in reversed(subexprs)] +
                              [r'\Z']))


def tokens_to_re(tokens):
    """Convert tokens to a regular expression for exact matching."""

    def token_to_re(token):
        """Convert tokens to simple regular expressions."""
        if token == '*':
            return r'.*'
        else:
            return re.escape(token)

    return re.compile("".join(map(token_to_re, tokens) + [r'\Z']))


class SimplePattern(Pattern):
    """Pattern that matches according to the given pattern expression."""

    def __init__(self, pattern):
        """Initialize an instance."""
        super(SimplePattern, self).__init__()
        tokens = PATTERN_TOKEN_RE.findall(pattern)
        self.partial_re = tokens_to_partial_re(tokens)
        self.full_re = tokens_to_re(tokens)

    def prefix_matches(self, partial_name):
        """Return True if the partial name matches."""
        return self.partial_re.match(partial_name) is not None

    def matches(self, name):
        """Return True if the complete name matches."""
        return self.full_re.match(name) is not None


class PrefixPattern(Pattern):
    """Pattern that assumes a previously given prefix."""

    def __init__(self, prefix, pattern):
        """Initialize an instance."""
        super(PrefixPattern, self).__init__()
        self.prefix = prefix
        self.pattern = pattern

    def assume_prefix(self, prefix):
        """Return an equivalent pattern with the given prefix baked in."""
        return PrefixPattern(self.prefix + prefix, self.pattern)

    def prefix_matches(self, partial_name):
        """Return True if the partial name matches."""
        return self.pattern.prefix_matches(self.prefix + partial_name)

    def matches(self, name):
        """Return True if the complete name matches."""
        return self.pattern.matches(self.prefix + name)


class SumPattern(Pattern):
    """Pattern that matches if at least one given pattern matches."""

    def __init__(self, patterns):
        """Initialize an instance."""
        super(SumPattern, self).__init__()
        self.patterns = patterns

    def prefix_matches(self, partial_name):
        """Return True if the partial name matches."""
        for pattern in self.patterns:
            if pattern.prefix_matches(partial_name):
                return True
        return False

    def matches(self, name):
        """Return True if the complete name matches."""
        for pattern in self.patterns:
            if pattern.matches(name):
                return True
        return False



from argparse import ArgumentParser
from netaddr import IPNetwork
from socket import gethostbyname
import sys
import yaml


COMMANDS = {
    'aws': ('aws ec2 authorize-security-group-egress --group-id {group}'
            ' --protocol {protocol} --port {port} --cidr {to_cidr}'),
    'neutron': ('neutron security-group-rule-create --direction egress'
                ' --ethertype {ip_version} --protocol {protocol} '
                ' --port-range-min {port} --port-range-max {port} '
                ' --remote-ip-prefix {to_cidr} {group}'),
    'nova': ('nova secgroup-add-rule {group} {protocol} {port} {port}'
             ' {to_cidr}'),
    'iptables': ('iptables -A FORWARD -p {protocol} --dport {port}'
                 ' -s {from_host} -d {to_host} -j ACCEPT'),
    'ufw': ('ufw allow proto {protocol} from any to {to_cidr} port {port}'),
}

COMMAND_ALIASES = {
    'amazon': 'aws',
    'ec2': 'aws',
    'openstack': 'neutron',
    'os': 'neutron',
}

SECGROUP_COMMANDS = ('aws', 'neutron', 'nova')


def merge_yaml(paths, use_from=False):
    """Merge multiple firewall YAML rule files with hosts/ports de-duped."""
    merged_rules = {}

    for path in paths:
        rules = yaml.load(open(path))

        for rule in rules.get('egress', []):
            from_ip = IPNetwork(gethostbyname(rule['from_host']))
            to_ip = IPNetwork(gethostbyname(rule['to_host']))

            # We need these values for de-duping, so we may as well add them
            # into the rule here to use later when generating commands.
            rule['to_cidr'] = str(to_ip.cidr)
            rule['ip_version'] = to_ip.version

            if use_from:
                rule['from_cidr'] = str(from_ip.cidr)
                key_template = '{protocol}:{from_cidr}:{to_cidr}'
            else:
                key_template = '{protocol}:{to_cidr}'
            key = key_template.format(**rule)

            if key not in merged_rules:
                merged_rules[key] = rule
            else:
                ports = merged_rules[key]['ports']
                ports_diff = set(ports).difference(rule['ports'])

                if ports_diff:
                    merged_rules[key]['ports'] = ports + ports_diff

    return merged_rules.values()


def generate_commands(cmd, rules, group=None):
    """Generate firewall client commands from conn-check firewall rules."""
    output = []
    for rule in rules:
        for port in rule['ports']:
            params = {
                'group': group or '$SECGROUP',
                'port': port,
            }
            params.update(rule)
            output.append(COMMANDS[cmd].format(**params))

    return output


def run(*args):
    parser = ArgumentParser()
    parser.add_argument('-t', '--type', dest='output_type', required=True,
                        help="Rules output type, e.g. neutron, nova, aws,"
                        " iptables, ufw.")
    parser.add_argument("paths", nargs='+',
                        help="Paths to YAML files to combine/parse.")
    parser.add_argument('--group', dest='group', required=False,
                        help="AWS security group ID or OpenStack Neutron group"
                        " name.")
    options = parser.parse_args(list(args))

    available_commands = COMMANDS.keys() + COMMAND_ALIASES.keys()
    output_type = options.output_type.lower()
    if output_type not in available_commands:
        sys.stderr.write('Error: invalid output type ({})\n'.format(
                         options.output_type))
        return 1

    command_type = COMMAND_ALIASES.get(output_type, output_type)
    rules = merge_yaml(options.paths, (command_type in SECGROUP_COMMANDS))
    output_rules = generate_commands(command_type, rules, options.group)

    sys.stdout.write('{}\n'.format('\n'.join(output_rules)))
    return 0


def main():
    sys.exit(run(*sys.argv[1:]))


if __name__ == '__main__':
    main()



import sys
import socket
import yaml

from ..main import Command, parse_version_arg


class FirewallRulesOutput(object):
    """Outputs a set of YAML firewall rules matching checks."""

    def __init__(self, output):
        self.output = output
        self.output_data = {}
        self.fqdn = socket.getfqdn()

    def notify_skip(self, name):
        """Passes skips.

        Should pass everything when using the skipping_strategy, directly to
        write().
        """

        self.write(name)

    def write(self, data):
        """Filters out non-TCP/UDP checks.

        Stores host/port/proto info for output later as YAML.
        """

        # We only need TCP/UDP checks
        if not any(x in data for x in ('tcp', 'udp')):
            return

        # Here we take the list of colon separated values in reverse order, so
        # we should get just the host/port/proto for the check without the
        # specific prefix (e.g. memcache, http)
        port, host, protocol = reversed(data.split(':')[-3:])
        protocol = protocol.strip()

        key = "{}:{}".format(host, protocol)
        if key not in self.output_data:
            self.output_data[key] = {
                'from_host': self.fqdn,
                'to_host': host,
                'ports': [],
                'protocol': protocol,
            }

        port = int(port)
        if port not in self.output_data[key]['ports']:
            self.output_data[key]['ports'].append(port)

    def flush(self):
        """Outputs our structured egress firewall info as YAML."""

        self.output.write(yaml.dump({'egress': self.output_data.values()}))


class FirewallExportCommand(Command):
    """CLI command runner for conn-check-export-fw"""

    def wrap_output(self, output):
        """Wraps output stream.

        Override some options in order to just output fw rules without
        performing checks.
        """
        # We don't want to actually perform the checks
        self.options.dry_run = True
        self.options.buffer_output = False
        self.options.show_duration = False

        super(FirewallExportCommand, self).wrap_output(output)

        self.output = FirewallRulesOutput(self.output)
        self.results.output = self.output


def run(*args):
    if parse_version_arg():
        return 0

    cmd = FirewallExportCommand(args)
    return cmd.run()


def main():
    sys.exit(run(*sys.argv[1:]))


if __name__ == '__main__':
    main()



from email.mime.text import MIMEText
import glob
import os
from pkg_resources import resource_stream
from StringIO import StringIO
import urlparse

from OpenSSL import SSL
from OpenSSL.crypto import load_certificate, FILETYPE_PEM

from twisted.internet import reactor, ssl
from twisted.internet.error import DNSLookupError, TimeoutError
from twisted.internet.abstract import isIPAddress
from twisted.internet.defer import (
    Deferred,
    inlineCallbacks,
    )
from twisted.internet.protocol import (
    ClientCreator,
    DatagramProtocol,
    Protocol,
    )
from twisted.mail.smtp import ESMTPSenderFactory
from twisted.protocols.memcache import MemCacheProtocol

from txrequests import Session
from requests.auth import HTTPDigestAuth

try:
    from requests.packages.urllib3 import disable_warnings
    from requests.packages.urllib3.contrib.pyopenssl import inject_into_urllib3
except ImportError:
    from urllib3 import disable_warnings
    from urllib3.contrib.pyopenssl import inject_into_urllib3

from .check_impl import (
    add_check_prefix,
    make_check,
    sequential_check,
    )


# Ensure we always use pyOpenSSL instead of the ssl builtin
inject_into_urllib3()

CA_CERTS = []


def load_tls_certs(path):
    cert_map = {}
    for filepath in glob.glob("{}/*.pem".format(os.path.abspath(path))):
        # There might be some dead symlinks in there,
        # so let's make sure it's real.
        if os.path.isfile(filepath):
            data = open(filepath).read()
            x509 = load_certificate(FILETYPE_PEM, data)
            # Now, de-duplicate in case the same cert has multiple names.
            cert_map[x509.digest('sha1')] = x509

    CA_CERTS.extend(cert_map.values())


class TCPCheckProtocol(Protocol):

    def connectionMade(self):
        self.transport.loseConnection()


class VerifyingContextFactory(ssl.CertificateOptions):

    def __init__(self, verify, caCerts, verifyCallback=None):
        ssl.CertificateOptions.__init__(self, verify=verify,
                                        caCerts=caCerts,
                                        method=SSL.SSLv23_METHOD)
        self.verifyCallback = verifyCallback

    def _makeContext(self):
        context = ssl.CertificateOptions._makeContext(self)
        if self.verifyCallback is not None:
            context.set_verify(
                SSL.VERIFY_PEER | SSL.VERIFY_FAIL_IF_NO_PEER_CERT,
                self.verifyCallback)
        return context


@inlineCallbacks
def do_tcp_check(host, port, tls=False, tls_verify=True,
                 timeout=None):
    """Generic connection check function."""
    if not isIPAddress(host):
        try:
            ip = yield reactor.resolve(host, timeout=(1, timeout))
        except DNSLookupError:
            raise ValueError("dns resolution failed")
    else:
        ip = host
    creator = ClientCreator(reactor, TCPCheckProtocol)
    try:
        if tls:
            context = VerifyingContextFactory(tls_verify, CA_CERTS)
            yield creator.connectSSL(ip, port, context,
                                     timeout=timeout)
        else:
            yield creator.connectTCP(ip, port, timeout=timeout)
    except TimeoutError:
        if ip == host:
            raise ValueError("timed out")
        else:
            raise ValueError("timed out connecting to {}".format(ip))


def make_tcp_check(host, port, timeout=None, **kwargs):
    """Return a check for TCP connectivity."""
    return make_check("tcp:{}:{}".format(host, port),
                      lambda: do_tcp_check(host, port, timeout=timeout),
                      info="{}:{}".format(host, port))


def make_tls_check(host, port, disable_tls_verification=False, timeout=None,
                   **kwargs):
    """Return a check for TLS setup."""

    verify = not disable_tls_verification
    check = make_check("tls:{}:{}".format(host, port),
                       lambda: do_tcp_check(host, port, tls=True,
                                            tls_verify=verify,
                                            timeout=timeout),
                       info="{}:{}".format(host, port))

    return check


class UDPCheckProtocol(DatagramProtocol):

    def __init__(self, host, port, send, expect, deferred=None,
                 timeout=None):
        self.host = host
        self.port = port
        self.send = send
        self.expect = expect
        self.deferred = deferred
        self.timeout = timeout

    def _finish(self, success, result):
        if not (self.delayed.cancelled or self.delayed.called):
            self.delayed.cancel()
        if self.deferred is not None:
            if success:
                self.deferred.callback(result)
            else:
                self.deferred.errback(result)
            self.deferred = None

    def startProtocol(self):
        self.transport.write(self.send, (self.host, self.port))
        self.delayed = reactor.callLater(self.timeout,
                                         self._finish,
                                         False, TimeoutError())

    def datagramReceived(self, datagram, addr):
        if datagram == self.expect:
            self._finish(True, True)
        else:
            self._finish(False, ValueError("unexpected reply"))


@inlineCallbacks
def do_udp_check(host, port, send, expect, timeout=None):
    """Generic connection check function."""
    if not isIPAddress(host):
        try:
            ip = yield reactor.resolve(host, timeout=(1, timeout))
        except DNSLookupError:
            raise ValueError("dns resolution failed")
    else:
        ip = host
    deferred = Deferred()
    protocol = UDPCheckProtocol(ip, port, send, expect, deferred, timeout)
    reactor.listenUDP(0, protocol)
    try:
        yield deferred
    except TimeoutError:
        if ip == host:
            raise ValueError("timed out")
        else:
            raise ValueError("timed out waiting for {}".format(ip))


def make_udp_check(host, port, send, expect, timeout=None,
                   **kwargs):
    """Return a check for UDP connectivity."""
    return make_check("udp:{}:{}".format(host, port),
                      lambda: do_udp_check(host, port, send, expect, timeout),
                      info="{}:{}".format(host, port))


def extract_host_port(url):
    parsed = urlparse.urlparse(url)
    host = parsed.hostname
    port = parsed.port
    scheme = parsed.scheme
    if not scheme:
        scheme = 'http'
    if port is None:
        if scheme == 'https':
            port = 443
        else:
            port = 80
    return host, port, scheme


def make_http_check(url, method='GET', expected_code=200, **kwargs):
    subchecks = []
    host, port, scheme = extract_host_port(url)
    proxy_url = kwargs.get('proxy_url')
    proxy_host = kwargs.get('proxy_host')
    proxy_port = int(kwargs.get('proxy_port', 8000))
    timeout = kwargs.get('timeout', None)
    expected_code = int(expected_code)

    if proxy_host:
        subchecks.append(make_tcp_check(proxy_host, proxy_port,
                                        timeout=timeout))
    else:
        subchecks.append(make_tcp_check(host, port, timeout=timeout))

    @inlineCallbacks
    def do_request():
        proxies = {}
        if proxy_url:
            proxies['http'] = proxies['https'] = proxy_url
        elif proxy_host:
            proxies['http'] = proxies['https'] = '{}:{}'.format(
                proxy_host, proxy_port)

        headers = kwargs.get('headers')
        body = kwargs.get('body')
        disable_tls_verification = kwargs.get('disable_tls_verification',
                                              False)
        allow_redirects = kwargs.get('allow_redirects', False)
        params = kwargs.get('params')
        cookies = kwargs.get('cookies')
        auth = kwargs.get('auth')
        digest_auth = kwargs.get('digest_auth')

        args = {
            'method': method,
            'url': url,
            'verify': not disable_tls_verification,
            'timeout': timeout,
            'allow_redirects': allow_redirects,
        }
        if headers:
            args['headers'] = headers
        if body:
            args['data'] = body
        if proxies:
            args['proxies'] = proxies
        if params:
            args['params'] = params
        if cookies:
            args['cookies'] = cookies
        if auth:
            args['auth'] = auth
        if digest_auth:
            args['auth'] = HTTPDigestAuth(digest_auth)

        if disable_tls_verification:
            disable_warnings()

        with Session() as session:
            request = session.request(**args)

            response = yield request
            if response.status_code != expected_code:
                raise RuntimeError(
                    "Unexpected response code: {}".format(
                        response.status_code))

    subchecks.append(make_check('', do_request,
                     info='{} {}'.format(method, url)))

    return add_check_prefix('http:{}'.format(url),
                            sequential_check(subchecks))


def make_amqp_check(host, port, username, password, use_tls=True, vhost="/",
                    timeout=None, **kwargs):
    """Return a check for AMQP connectivity."""
    from txamqp.protocol import AMQClient
    from txamqp.client import TwistedDelegate
    from txamqp.spec import load as load_spec

    subchecks = []
    subchecks.append(make_tcp_check(host, port, timeout=timeout))

    if use_tls:
        subchecks.append(make_tls_check(host, port, verify=False,
                                        timeout=timeout))

    @inlineCallbacks
    def do_auth():
        """Connect and authenticate."""
        delegate = TwistedDelegate()
        spec = load_spec(resource_stream('conn_check', 'amqp0-8.xml'))
        creator = ClientCreator(reactor, AMQClient,
                                delegate, vhost, spec)
        client = yield creator.connectTCP(host, port, timeout=timeout)
        yield client.authenticate(username, password)

    subchecks.append(make_check("amqp:{}:{}".format(host, port),
                                do_auth, info="user {}".format(username),))
    return sequential_check(subchecks)


def make_smtp_check(host, port, username, password, from_address, to_address,
                    message='', subject='', helo_fallback=False, use_tls=True,
                    timeout=None, **kwargs):
    """Return a check for SMTP connectivity."""

    subchecks = []
    subchecks.append(make_tcp_check(host, port, timeout=timeout))

    if use_tls:
        subchecks.append(make_tls_check(host, port, verify=False,
                                        timeout=timeout))

    @inlineCallbacks
    def do_connect():
        """Connect and authenticate."""
        result_deferred = Deferred()
        context_factory = None
        if use_tls:
            from twisted.internet import ssl as twisted_ssl
            context_factory = twisted_ssl.ClientContextFactory()

        body = MIMEText(message)
        body['Subject'] = subject
        factory = ESMTPSenderFactory(
            username,
            password,
            from_address,
            to_address,
            StringIO(body.as_string()),
            result_deferred,
            contextFactory=context_factory,
            requireTransportSecurity=use_tls,
            requireAuthentication=True,
            heloFallback=helo_fallback)

        if use_tls:
            reactor.connectSSL(host, port, factory, context_factory)
        else:
            reactor.connectTCP(host, port, factory)
        result = yield result_deferred

        if result[0] == 0:
            raise RuntimeError("failed to send email via smtp")

    subchecks.append(make_check("smtp:{}:{}".format(host, port),
                                do_connect, info="user {}".format(username),))
    return sequential_check(subchecks)


def make_postgres_check(host, port, username, password, database,
                        timeout=None, **kwargs):
    """Return a check for Postgres connectivity."""

    import psycopg2
    subchecks = []
    connect_kw = {
        'host': host,
        'user': username,
        'database': database,
        'connect_timeout': timeout,
    }

    if host[0] != '/':
        connect_kw['port'] = port
        subchecks.append(make_tcp_check(host, port, timeout=timeout))

    if password is not None:
        connect_kw['password'] = password

    def check_auth():
        """Try to establish a postgres connection and log in."""
        conn = psycopg2.connect(**connect_kw)
        conn.close()

    subchecks.append(make_check("postgres:{}:{}".format(host, port),
                                check_auth, info="user {}".format(username),
                                blocking=True))

    return sequential_check(subchecks)


def make_redis_check(host, port, password=None, timeout=None,
                     **kwargs):
    """Make a check for the configured redis server."""
    import txredis
    subchecks = []
    subchecks.append(make_tcp_check(host, port, timeout=timeout))

    @inlineCallbacks
    def do_connect():
        """Connect and authenticate.
        """
        client_creator = ClientCreator(reactor, txredis.client.RedisClient)
        client = yield client_creator.connectTCP(host=host, port=port,
                                                 timeout=timeout)

        if password is None:
            ping = yield client.ping()
            if not ping:
                raise RuntimeError("failed to ping redis")
        else:
            resp = yield client.auth(password)
            if resp != 'OK':
                raise RuntimeError("failed to auth to redis")

    if password is not None:
        connect_info = "connect with auth"
    else:
        connect_info = "connect"
    subchecks.append(make_check(connect_info, do_connect))

    return add_check_prefix('redis:{}:{}'.format(host, port),
                            sequential_check(subchecks))


def make_memcache_check(host, port, password=None, timeout=None,
                        **kwargs):
    """Make a check for the configured redis server."""
    subchecks = []
    subchecks.append(make_tcp_check(host, port, timeout=timeout))

    @inlineCallbacks
    def do_connect():
        """Connect and authenticate."""
        client_creator = ClientCreator(reactor, MemCacheProtocol)
        client = yield client_creator.connectTCP(host=host, port=port,
                                                 timeout=timeout)

        version = yield client.version()
        if version is None:
            raise RuntimeError('Failed retrieve memcached server version')

    subchecks.append(make_check('connect', do_connect))

    return add_check_prefix('memcache:{}:{}'.format(host, port),
                            sequential_check(subchecks))


def make_mongodb_check(host, port=27017, username=None, password=None,
                       database='test', timeout=10, **kwargs):
    """Return a check for MongoDB connectivity."""

    import txmongo
    subchecks = []
    subchecks.append(make_tcp_check(host, port, timeout=timeout))

    port = int(port)

    @inlineCallbacks
    def do_connect():
        """Try to establish a mongodb connection."""
        conn = txmongo.MongoConnection(host, port)

        conn.uri['options']['connectTimeoutMS'] = int(timeout*1000)
        if username:
            conn.uri['username'] = username
        if password:
            conn.uri['password'] = password

        # We don't start our timeout callback until now, otherwise we might
        # elapse part of our timeout period during the earlier TCP check
        reactor.callLater(timeout, timeout_handler)

        mongo = yield conn
        names = yield mongo[database].collection_names()
        if names is None:
            raise RuntimeError('Failed to retrieve collection names')

    def timeout_handler():
        """Manual timeout handler as txmongo timeout args don't always work."""
        if 'deferred' in do_connect.func_dict:
            err = ValueError("timeout connecting to mongodb")
            do_connect.func_dict['deferred'].errback(err)

    if any((username, password)):
        connect_info = "connect with auth"
    else:
        connect_info = "connect"
    subchecks.append(make_check(connect_info, do_connect))

    return add_check_prefix('mongodb:{}:{}'.format(host, port),
                            sequential_check(subchecks))


CHECKS = {
    'tcp': {
        'fn': make_tcp_check,
        'args': ['host', 'port'],
    },
    'tls': {
        'fn': make_tls_check,
        'args': ['host', 'port'],
    },
    'udp': {
        'fn': make_udp_check,
        'args': ['host', 'port', 'send', 'expect'],
    },
    'http': {
        'fn': make_http_check,
        'args': ['url'],
    },
    'amqp': {
        'fn': make_amqp_check,
        'args': ['host', 'port', 'username', 'password'],
    },
    'postgres': {
        'fn': make_postgres_check,
        'args': ['host', 'port', 'username', 'password', 'database'],
    },
    'redis': {
        'fn': make_redis_check,
        'args': ['host', 'port'],
    },
    'memcache': {
        'fn': make_memcache_check,
        'args': ['host', 'port'],
    },
    'mongodb': {
        'fn': make_mongodb_check,
        'args': ['host'],
    },
    'smtp': {
        'fn': make_smtp_check,
        'args': ['host', 'port', 'username', 'password'],
    },
}

CHECK_ALIASES = {
    'mongo': 'mongodb',
    'memcached': 'memcache',
    'ssl': 'tls',
    'postgresql': 'postgres',
    'https': 'http',
}
